import csv
import json as jsonlib
import pickle
from tabulate import tabulate  # https://pypi.org/project/tabulate/
import asciichartpy as chart
from math import nan

from spyql.log import user_error
from spyql.nulltype import NULL


class Writer:
    @staticmethod
    def make_writer(writer_name, outputfile, options):
        try:
            if not writer_name:
                return CSVWriter(outputfile, **options)
            writer_name = writer_name.upper()
            if writer_name == "CSV":
                return CSVWriter(outputfile, **options)
            if writer_name == "JSON":
                return SimpleJSONWriter(outputfile, **options)
            if writer_name == "PRETTY":
                return PrettyWriter(outputfile, **options)
            if writer_name == "SPY":
                return SpyWriter(outputfile, **options)
            if writer_name == "SQL":
                return SQLWriter(outputfile, **options)
            if writer_name == "PLOT":
                return PlotWriter(outputfile, **options)
        except TypeError as e:
            user_error(f"Could not create '{writer_name}' writer", e)
        user_error(
            f"Unknown writer '{writer_name}'",
            SyntaxError("Error parsing TO statement"),
            writer_name,
        )

    def __init__(self, outputfile):
        self.outputfile = outputfile

    def writeheader(self, header):
        self.header = header

    def writerow(self, row):
        raise NotImplementedError

    def writerows(self, rows):
        for r in rows:
            self.writerow(r)

    def flush(self):
        pass


class CSVWriter(Writer):
    def __init__(self, outputfile, header=True, **options):
        super().__init__(outputfile)
        self.header_on = header
        self.csv = csv.writer(outputfile, **options)

    def writeheader(self, header):
        if self.header_on:
            self.csv.writerow(header)

    def writerow(self, row):
        self.csv.writerow(row)

    def writerows(self, rows):
        self.csv.writerows(rows)


class SimpleJSONWriter(Writer):
    def __init__(self, outputfile, **options):
        super().__init__(outputfile)
        jsonlib.dumps({"a": 1}, **options)  # test options
        self.options = options

    def writerow(self, row):
        self.outputfile.write(self.makerow(row) + "\n")

    def makerow(self, row):
        single_dict = (
            self.header in [["col1"], ["json"]]
            and len(row) == 1
            and isinstance(row[0], dict)
        )
        obj = row[0] if single_dict else dict(zip(self.header, row))
        return jsonlib.dumps(
            obj, default=lambda x: None if x is NULL else str(x), **self.options
        )


class CollectWriter(Writer):
    """
    Abstract writer that collects all records into a (in-memory) list and dumps all
    the output records at the end.
    Child classes must implement the `writerows` method.
    """

    def __init__(self, outputfile):
        super().__init__(outputfile)
        self.all_rows = []  # needs to store output in memory

    def transformvalue(self, value):
        return None if value is NULL else value

    def writerow(self, row):
        self.all_rows.append([self.transformvalue(val) for val in row])  # accumulates

    def writerows(self, rows):
        raise NotImplementedError

    def flush(self):
        if self.all_rows:
            self.writerows(self.all_rows)  # dumps


class PrettyWriter(CollectWriter):
    def __init__(self, outputfile, header=True, **options):
        super().__init__(outputfile)
        tabulate([[1, 2, 3]], **options)  # test options
        self.header_on = header
        self.options = options

    def writerows(self, rows):
        # TODO handle default tablefmt
        self.outputfile.write(
            tabulate(
                rows,
                self.header if self.header_on else [],
                **self.options,
            )
        )
        self.outputfile.write("\n")


class PlotWriter(CollectWriter):
    def __init__(self, outputfile, header=True, height=20):
        super().__init__(outputfile)
        self.header_on = header
        self.height = height

    def transformvalue(self, value):
        return nan if value is NULL or value is None else value

    def writerows(self, rows):
        colors = [
            chart.cyan,
            chart.red,
            chart.magenta,
            chart.lightgray,
            chart.green,
            chart.blue,
        ]
        config = {"height": self.height, "colors": colors}

        # first transpose rows into cols
        cols = list(map(list, zip(*rows)))

        self.outputfile.write(chart.plot(cols, config))
        if self.header and self.header_on:
            self.outputfile.write("\n\nLegend: ")
            for i in range(len(self.header)):
                self.outputfile.write(
                    "\t"
                    + colors[i % len(colors)]
                    + "─── "
                    + self.header[i]
                    + chart.reset
                    + " "
                )
            self.outputfile.write("\n")


class SpyWriter(Writer):
    def __init__(self, outputfile):
        super().__init__(outputfile)

    @staticmethod
    def pack(row):
        return pickle.dumps(row).hex() + "\n"

    def writeheader(self, header):
        # TODO first line is a dict with list of cols, version, etc
        self.outputfile.write(self.pack(header))

    def writerow(self, row):
        self.outputfile.write(self.pack(row))


class SQLWriter(Writer):
    def __init__(self, outputfile, chunk_size=1000, table="table_name"):
        super().__init__(outputfile)
        self.chunk_size = chunk_size
        self.table_name = table
        self.chunk = []

    def writeheader(self, header):
        self.statement = (
            f'INSERT INTO "{self.table_name}"('
            + ",".join(['"{}"'.format(h) for h in header])
            + ") VALUES {};\n"
        )

    def writerow(self, row):
        self.chunk.append(
            "({})".format(
                ",".join(
                    [
                        str(v)
                        if isinstance(v, int) or isinstance(v, float)
                        else (
                            "NULL"
                            if v is NULL or v is None
                            else "'{}'".format(str(v).replace("'", "''"))
                        )
                        for v in row
                    ]
                )
            )
        )

        if len(self.chunk) >= self.chunk_size:
            self.writestatement()

    def writestatement(self):
        self.outputfile.write(self.statement.format(",".join(self.chunk)))
        self.chunk = []

    def flush(self):
        if self.chunk:  # write leftovers...
            self.writestatement()



import re


def quote_ifstr(s):
    return f"'{s}'" if isinstance(s, str) else s


def make_str_valid_varname(s):
    # remove invalid characters (except spaces in-between)
    s = re.sub(r"[^0-9a-zA-Z_\s]", " ", s).strip()

    # replace spaces by underscores (instead of dropping spaces) for readability
    s = re.sub(r"\s+", "_", s)

    # if first char is not letter or underscore then add underscore to make it valid
    if not re.match("^[a-zA-Z_]", s):
        s = "_" + s

    return s


def try2eval(val, globals={}, locals={}):
    try:
        return eval(val, globals, locals)
    except Exception:
        return val


def isiterable(x):
    """Returns True if `x` is iterable that is not a string and is not null"""
    from spyql.nulltype import Null
    from collections.abc import Iterable

    return isinstance(x, Iterable) and x is not Null and not isinstance(x, str)



# Defines the NULL datatype that mimics SQL's NULL
# Unlike None, which throws exceptions, operations with NULLs return NULL
# Allows getting items so that `a.get('b', NULL).get('c', NULL)`` returns NULL
#  when `b` does not exist, and `NULL[x]` returns NULL

import spyql.log


class NullType:
    def __new__(cls):
        return NULL

    def __reduce__(self):
        return (NullType, ())

    def __copy__(self):
        return NULL

    def __deepcopy__(self, a):
        return NULL

    def __call__(self, default):
        pass

    def __repr__(self):
        return "NULL"

    def __str__(self):
        return ""

    def __hash__(self):
        return hash("NULL")

    def __bool__(self):
        return False

    def __lt__(self, other):
        return self

    def __le__(self, other):
        return self

    def __eq__(self, other):
        return self

    def __ne__(self, other):
        return self

    def __ge__(self, other):
        return self

    def __gt__(self, other):
        return self

    def __abs__(self):
        return self

    def __add__(self, other):
        return self

    def __and__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __invert__(self):
        return self

    def __lshift__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __matmul__(self, other):
        return self

    def __neg__(self):
        return self

    def __or__(self, other):
        return self

    def __pos__(self):
        return self

    def __pow__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __contains__(self, other):
        return False

    def __delitem__(self, other):
        pass

    def __getitem__(self, other):
        return self

    def __setitem__(self, other, val):
        pass

    def __radd__(self, other):
        return self

    def __rand__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __rlshift__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __rmatmul__(self, other):
        return self

    def __ror__(self, other):
        return self

    def __rpow__(self, other):
        return self

    def __rrshift__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __rxor__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __iand__(self, other):
        return self

    def __ifloordiv__(self, other):
        return self

    def __ilshift__(self, other):
        return self

    def __imod__(self, other):
        return self

    def __imul__(self, other):
        return self

    __array_priority__ = 10000

    def __imatmul__(self, other):
        return self

    def __ior__(self, other):
        return self

    def __ipow__(self, other):
        return self

    def __irshift__(self, other):
        return self

    def __isub__(self, other):
        return self

    def __itruediv__(self, other):
        return self

    def __ixor__(self, other):
        return self

    def __len__(self):
        return 0

    def __iter__(self):
        return [].__iter__()

    def __round__(self, ndigits=0):
        return self

    def __trunc__(self):
        return self

    def __floor__(self):
        return self

    def __ceil__(self):
        return self

    def get(self, *args, **kwargs):
        return self


# singleton
try:
    NULL
    # explanation here:
    # https://stackoverflow.com/questions/41048643/how-to-create-a-second-none-in-python-making-a-singleton-object-where-the-id-is
except NameError:
    NULL = object.__new__(NullType)
    Null = NULL  # alias
    null = NULL  # alias

# functions that support NULLs (and that need to be replaced in the query)
NULL_SAFE_FUNCS = {
    "int": "int_",
    "float": "float_",
    "str": "str_",
    "complex": "complex_",
}


class NullSafeDict(dict):
    __slots__ = ()  # no __dict__

    @staticmethod
    def none2null(adic):
        def none2null_el(el):
            return [NULL if x is None else x for x in el] if type(el) is list else el

        return {k: NULL if v is None else none2null_el(v) for k, v in adic.items()}

    def __init__(self, adic, **kwargs):
        super().__init__(
            # converts None -> NULL
            NullSafeDict.none2null(adic),
            **kwargs
        )

    # returns NULL when key is not found
    def __missing__(self, key):
        spyql.log.user_warning4func("key not found", KeyError(key), key)
        return NULL

    def __hash__(self):
        # TODO make dict immutable
        import json

        # TODO check if this is sufficienly efficient...
        # This only needs to guarantee that two equivalent dicts have the same hash
        return hash(json.dumps(self, default=lambda x: str(x), sort_keys=True))


# returns default if val is NULL otherwise returns val
def coalesce(val, default):
    if val is NULL:
        return default
    return val


ifnull = coalesce  # alias


# returns NULL if a equals b otherwise returns a
def nullif(a, b):
    if a == b:
        return NULL
    return a


# returns NULL if any argument equals NULL
def null_safe_call(fun, *args, **kwargs):
    if NULL in args or NULL in kwargs.values():
        return NULL
    return fun(*args, **kwargs)


# NULL-safe functions
def float_(a):
    if a is NULL:
        return NULL
    try:
        return float(a)
    except ValueError as e:
        spyql.log.conversion_warning("float", e, a)
        return NULL


def int_(a, *args, **kwargs):
    if a is NULL or NULL in args or NULL in kwargs.values():
        return NULL
    try:
        return int(a, *args, **kwargs)
    except ValueError as e:
        spyql.log.conversion_warning("int", e, a, **kwargs)
        return NULL


def complex_(*args):
    if NULL in args:
        return NULL
    try:
        return complex(*args)
    except ValueError as e:
        spyql.log.conversion_warning("complex", e, *args)
        return NULL


def str_(*args, **kwargs):
    if NULL in args or NULL in kwargs.values():
        return NULL
    return str(*args, **kwargs)



import csv
import json as jsonlib
import pickle
import sys
import io
import re
import os
from itertools import islice, chain
from io import StringIO
import copy

from spyql.writer import Writer
from spyql.output_handler import OutputHandler
import spyql.nulltype
import spyql.log
from spyql.utils import make_str_valid_varname, isiterable
import spyql.agg


def init_vars():
    """Initializes dict of variables for user queries"""
    vars = dict()
    # imports for user queries (TODO move to init.py when mature)
    exec(
        "from datetime import datetime, date, timezone\n"
        "from spyql.nulltype import *\n"
        "from spyql.agg import *\n"
        "from math import *\n"
        "import re\n",
        {},
        vars,
    )

    try:
        # user defined imports, functions, etc
        config_home = os.environ.get(
            "XDG_CONFIG_HOME", os.path.expanduser(os.path.join("~", ".config"))
        )
        init_fname = os.path.join(config_home, "spyql", "init.py")
        with open(init_fname) as f:
            exec(f.read(), {}, vars)
            spyql.log.user_debug(f"Succesfully loaded {init_fname}")
    except FileNotFoundError:
        spyql.log.user_debug(f"Init file not found: {init_fname}")
    except Exception as e:
        spyql.log.user_warning(f"Could not load {init_fname}", e)
    return vars


class Processor:
    @staticmethod
    def make_processor(prs, strings, input_options):
        """
        Factory for making a file processor based on the parsed query
        """
        try:
            processor_name = prs["from"]
            if not processor_name:
                return Processor(prs, strings, **input_options)

            processor_name = processor_name.upper()

            if processor_name == "JSON":
                return JSONProcessor(prs, strings, **input_options)
            if processor_name == "CSV":
                return CSVProcessor(prs, strings, **input_options)
            if processor_name == "TEXT":  # single col
                return TextProcessor(prs, strings, **input_options)
            if processor_name == "SPY":
                return SpyProcessor(prs, strings, **input_options)

            return PythonExprProcessor(prs, strings, **input_options)
        except TypeError as e:
            spyql.log.user_error(f"Could not create '{processor_name}' processor", e)

    def __init__(self, prs, strings):
        self.prs = prs  # parsed query
        self.strings = strings  # quoted strings
        self.input_col_names = []  # column names of the input data
        self.translations = copy.deepcopy(
            spyql.nulltype.NULL_SAFE_FUNCS
        )  # map for alias, functions to be renamed...
        self.has_header = False
        self.casts = dict()
        self.col_values_exprs = []

    def reading_data(self):
        """
        Returns True after reading header, metadata, etc in input file
        """
        return True

    def handle_header_row(self, row):
        """
        Action for header row (e.g. column name definition)
        """
        pass

    def handle_1st_data_row(self, row):
        """
        Action for handling the first row of data
        """
        self.n_input_cols = len(row) if row else 0

        default_col_names = [
            self.default_col_name(_i) for _i in range(self.n_input_cols)
        ]
        self.col_values_exprs = [
            f"{self.casts[_i]}_(_values[{_i}])"
            if _i in self.casts
            else f"_values[{_i}]"
            for _i in range(self.n_input_cols)
        ]
        # dictionary to translate col names to accesses to `_values`
        self.translations.update(dict(zip(default_col_names, self.col_values_exprs)))
        if self.input_col_names:
            # TODO check if len(input_col_names) == self.n_input_cols
            self.translations.update(
                dict(zip(self.input_col_names, self.col_values_exprs))
            )

        # metadata: list of column names
        self.vars["_names"] = (
            self.input_col_names if self.input_col_names else default_col_names
        )
        # list of [col1,col2,...]
        cols_expr = "[" + ",".join(self.col_values_exprs) + "]"
        self.translations["cols"] = cols_expr
        # dict of {col1: value1, ...}
        self.translations["row"] = f"NullSafeDict(dict(zip(_names, {cols_expr})))"

    def make_out_cols_names(self, out_cols_names):
        """
        Creates list of output column names
        """
        input_col_names = self.input_col_names
        if not input_col_names:
            input_col_names = [
                self.default_col_name(i) for i in range(self.n_input_cols)
            ]
        out_cols_names = [
            (input_col_names if name == "*" else [name]) for name in out_cols_names
        ]
        out_cols_names = [
            name for sublist in out_cols_names for name in sublist
        ]  # flatten
        return out_cols_names

    def get_input_iterator(self):
        """
        Returns iterator over input (e.g. list if rows)
        Each row is list with one value per column
        e.g.
            [[1] ,[2], [3]]:                3 rows with a single col
            [[1,'a'], [2,'b'], [3,'c']]:    3 rows with 2 cols
        """
        return [[None]]  # default: returns a single line with a 'null' column

    def default_col_name(self, idx):
        """
        Default column names, e.g. col1 for the first column
        """
        return f"col{idx+1}"

    def prepare_expression(self, expr):
        """
        Replaces identifiers (column names) in sql expressions by references to
        `_values` and put (quoted) strings back
        """
        if expr == "*":
            return self.col_values_exprs

        if isinstance(expr, int):
            # special case: expression is out col number (1-based)
            return [f"_res[{expr-1}]"]  # reuses existing result

        for id, replacement in self.translations.items():
            pattern = rf"\b({id})\b"
            expr = re.compile(pattern).sub(replacement, expr)

        return [self.strings.put_strings_back(expr)]

    def is_clause_single(self, clause):
        """
        True if clause can only have a single expression
        """
        return clause not in {
            "select",
            "group by",
            "order by",
        }

    def compile_clause(self, clause, clause_modifier=None, mode="eval"):
        """
        Compiles a clause of the query
        """
        prs_clause = self.prs[clause]
        if not prs_clause:
            return None  # empty clause

        if clause_modifier:
            prs_clause = clause_modifier.format(prs_clause)

        single = self.is_clause_single(clause)
        clause_exprs = None
        if single:  # a clause with a single expression like WHERE
            clause_exprs = self.prepare_expression(prs_clause)
            if len(clause_exprs) > 1:
                spyql.log.user_error(
                    f"could not compile {clause.upper()} clause",
                    SyntaxError(
                        f"{clause.upper()} clause should not have more than 1"
                        " expression"
                    ),
                )
            clause_exprs = clause_exprs[0]
        else:  # a clause with multiple expressions like SELECT
            clause_exprs = [self.prepare_expression(c["expr"]) for c in prs_clause]
            clause_exprs = [
                item for sublist in clause_exprs for item in sublist
            ]  # flatten (because of '*')
            clause_exprs = ",".join(clause_exprs) + ","  # tuple constructor

        try:
            return compile(clause_exprs, f"<{clause}>", mode)
        except Exception as main_exception:
            if not single:
                # breaks down clause into expressions and tries
                # compiling one by one to detect in which expression
                # the error happened
                for c in range(len(prs_clause)):
                    try:
                        expr = prs_clause[c]["expr"]
                        translation = self.prepare_expression(expr)
                        for trans in translation:
                            if not trans.strip():
                                raise SyntaxError("empty expression")
                            compile(trans, f"<{clause}>", mode)
                    except Exception as expr_exception:
                        spyql.log.user_error(
                            f"could not compile {clause.upper()} expression #{c+1}",
                            expr_exception,
                            self.strings.put_strings_back(expr),
                        )

            spyql.log.user_error(
                f"could not compile {clause.upper()} clause", main_exception
            )

    def eval_clause(self, clause, clause_exprs, mode="eval"):
        """
        Evaluates/executes a previously compiled clause
        """
        if not clause_exprs:
            return
        cmd = eval if mode == "eval" else exec
        try:
            return cmd(clause_exprs, self.vars, self.vars)
        except Exception as main_exception:
            prs_clause = self.prs[clause]
            if not self.is_clause_single(clause):
                # breaks down clause into expressions and tries
                # evaluating/executing one by one to detect
                # in which expression the error happened
                for c in range(len(prs_clause)):
                    try:
                        expr = prs_clause[c]["expr"]
                        translation = self.prepare_expression(expr)
                        for trans in translation:
                            cmd(trans, self.vars, self.vars)
                    except Exception as expr_exception:
                        spyql.log.user_error(
                            f"could not evaluate {clause.upper()} expression #{c+1}",
                            expr_exception,
                            self.strings.put_strings_back(expr),
                            self.vars,
                        )

            spyql.log.user_error(
                f"could not evaluate {clause.upper()} clause",
                main_exception,
                vars=self.vars,
            )

    # main
    def go(self, output_file, output_options):
        output_handler = OutputHandler.make_handler(self.prs)
        writer = Writer.make_writer(self.prs["to"], output_file, output_options)
        output_handler.set_writer(writer)
        nrows_in = self._go(output_handler)
        output_handler.finish()
        spyql.log.user_info("#rows  in", nrows_in)
        spyql.log.user_info("#rows out", output_handler.rows_written)

    def _go(self, output_handler):
        select_expr = None
        where_expr = None
        explode_expr = None
        explode_cmd = None
        explode_its = [None]  # 1 element by default (no explosion)
        groupby_expr = None
        orderby_expr = None
        _values = []
        _res = tuple()
        _group_res = tuple()
        _sort_res = tuple()
        row_number = 0
        input_row_number = 0

        self.vars = init_vars()
        spyql.agg._init_aggs()

        # import user modules
        self.eval_clause(
            "import",
            self.compile_clause("import", "import {}", mode="exec"),
            mode="exec",
        )

        # gets user-defined output cols names (with AS alias)
        out_cols_names = [c["name"] for c in self.prs["select"]]

        # should not accept more than 1 source, joins, etc (at least for now)
        for _values in self.get_input_iterator():
            input_row_number = input_row_number + 1

            self.vars["_values"] = _values
            self.vars["input_row_number"] = input_row_number

            if not self.reading_data():
                self.handle_header_row(_values)
                continue

            # print header
            if select_expr is None:  # 1st input data row
                self.handle_1st_data_row(_values)
                output_handler.writer.writeheader(
                    self.make_out_cols_names(out_cols_names)
                )
                if output_handler.is_done():
                    return 0  # in case of `limit 0`

                select_expr = self.compile_clause("select")
                where_expr = self.compile_clause("where")
                explode_expr = self.compile_clause("explode")
                groupby_expr = self.compile_clause("group by")
                orderby_expr = self.compile_clause("order by")

            if explode_expr:
                explode_its = self.eval_clause("explode", explode_expr)

            for explode_it in explode_its:
                if explode_expr:
                    self.vars["explode_it"] = explode_it
                    if not explode_cmd:
                        explode_cmd = self.compile_clause(
                            "explode", "{} = explode_it", mode="exec"
                        )
                    self.eval_clause("explode", explode_cmd, mode="exec")
                    self.vars["_values"] = _values

                # filter (opt: eventually could be done before exploding)
                if not where_expr or self.eval_clause("where", where_expr):
                    # input line is eligible
                    row_number = row_number + 1
                    self.vars["row_number"] = row_number

                    if groupby_expr:
                        # group by can ref output columns, but does not depend on the
                        # execution of the select clause: refs to output columns are
                        # replaced by the correspondent expression
                        _group_res = self.eval_clause("group by", groupby_expr)
                        # we need to set the group key before running the select because
                        # aggregate functions need to know the group key beforehand
                        spyql.agg._start_new_agg_row(_group_res)

                    # calculate outputs
                    _res = self.eval_clause("select", select_expr)

                    if orderby_expr:
                        # in the order by clause, references to output columns use the
                        # outputs of the evaluation of the select expression
                        self.vars["_res"] = _res
                        _sort_res = self.eval_clause("order by", orderby_expr)

                    is_done = output_handler.handle_result(
                        _res, _sort_res, _group_res
                    )  # deal with output
                    if is_done:
                        # e.g. when reached limit
                        return input_row_number - (1 if self.has_header else 0)

        return input_row_number - (1 if self.has_header else 0)


class PythonExprProcessor(Processor):
    def __init__(self, prs, strings):
        super().__init__(prs, strings)

    # input is a Python expression
    def get_input_iterator(self):
        e = self.eval_clause("from", self.compile_clause("from"))
        if e:
            if not isiterable(e):
                e = [e]
            if not isiterable(e[0]):
                e = [[el] for el in e]
        return e


class TextProcessor(Processor):
    def __init__(self, prs, strings):
        super().__init__(prs, strings)

    # reads a text row as a row with 1 column
    def get_input_iterator(self):
        # to do: suport files
        return ([line.rstrip("\n\r")] for line in sys.stdin)


class SpyProcessor(Processor):
    def __init__(self, prs, strings):
        super().__init__(prs, strings)
        self.has_header = True

    def reading_data(self):
        return self.input_col_names

    def handle_header_row(self, row):
        self.input_col_names = row

    @staticmethod
    def unpack_line(line):
        return pickle.loads(bytes.fromhex(line))

    # input is a serialized Python list converted to hex
    def get_input_iterator(self):
        # to do: suport files
        return (self.unpack_line(line[0:-1]) for line in sys.stdin)


class JSONProcessor(Processor):
    def __init__(self, prs, strings, **options):
        super().__init__(prs, strings)
        jsonlib.loads('{"a": 1}', **options)  # test options
        self.options = options
        self.input_col_names = ["json"]

    # 1 row = 1 json
    def get_input_iterator(self):
        # TODO suport files

        # this might not be the most efficient way of converting None -> NULL, look at:
        # https://stackoverflow.com/questions/27695901/python-jsondecoder-custom-translation-of-null-type
        return (
            [
                jsonlib.loads(
                    line,
                    object_hook=lambda x: spyql.nulltype.NullSafeDict(x),
                    **self.options,
                )
            ]
            for line in sys.stdin
        )


# CSV
class CSVProcessor(Processor):
    def __init__(
        self, prs, strings, sample_size=10, header=None, infer_dtypes=True, **options
    ):
        super().__init__(prs, strings)
        self.sample_size = sample_size
        self.has_header = header
        self.infer_dtypes = infer_dtypes
        self.options = options
        csv.reader(StringIO("test"), **self.options)  # test options

    def _test_dtype(self, v):
        v = v.strip()
        if not v:
            return (-100, None)  # empty string: do not cast
        try:
            int(v)
            return (10, "int")
        except ValueError:
            try:
                float(v)
                return (20, "float")
            except ValueError:
                try:
                    complex(v)
                    return (30, "complex")
                except ValueError:
                    return (100, None)  # not a basic type: do not cast

    def _infer_dtypes(self, reader):
        if self.has_header:
            next(reader, None)  # skip header
        dtypes_rows = [[self._test_dtype(col) for col in line] for line in reader]
        if dtypes_rows and dtypes_rows[0]:
            dtypes = [
                max([row[c] for row in dtypes_rows]) for c in range(len(dtypes_rows[0]))
            ]
            for c in range(len(dtypes)):
                cast = dtypes[c][1]
                if cast:
                    self.casts[c] = cast

    def get_input_iterator(self):
        # Part 1 reads sample to detect dialect and if has header
        # TODO force linedelimiter to be new line char set

        # saves sample to a string
        # NOTE if dialect is given and type detection is off we should not need a sample
        sample = io.StringIO()
        for line in list(islice(sys.stdin, self.sample_size)):  # TODO: support files
            sample.write(line)
        sample_val = sample.getvalue()
        if not sample_val:
            return []
        if not self.options:
            # CSV dialect and header detection
            try:
                self.options = {"dialect": csv.Sniffer().sniff(sample_val)}
            except Exception as e:
                spyql.log.user_error("Could not detect CSV dialect from input", e)
            if self.has_header is None:
                try:
                    self.has_header = csv.Sniffer().has_header(sample_val)
                except Exception as e:
                    spyql.log.user_error("Could not detect if input CSV has header", e)
        elif self.has_header is None:
            self.has_header = True  # default if dialect is not automatically detected

        sample.seek(0)  # rewinds the sample
        if self.infer_dtypes:
            self._infer_dtypes(csv.reader(sample, **self.options))
            sample.seek(0)  # rewinds the sample again

        return chain(
            csv.reader(
                sample, **self.options
            ),  # goes through sample again (for reading input data)
            csv.reader(sys.stdin, **self.options),
        )  # continues to the rest of the file

    def reading_data(self):
        return (not self.has_header) or (self.input_col_names)

    def handle_header_row(self, row):
        self.input_col_names = [make_str_valid_varname(val) for val in row]



import logging
import asciichartpy as colors
import sys
import json

from spyql.utils import quote_ifstr

error_on_warning = False  # overrided by command line arg


def colored(val, color):
    """
    Creates string representation of `val` for color printing
    """
    return color + str(val) + colors.reset


def mk_user_msg(level, level_color, message, code=None, code_color=colors.lightyellow):
    return (
        colored(level, level_color)
        + f"\t{message}"
        + f"{': ' + colored(code, code_color) if code is not None else ''}"
    )


def user_error(message, exception, code=None, vars=None):
    """
    Reports an error, throwing the original exception
    Prints a custom message.
    Prints the data that originated the exception (if available).
    """
    sys.tracebacklimit = 0
    logging.error(mk_user_msg("ERROR", colors.lightred, message, code))
    if vars and vars.get("input_row_number") and vars.get("_values"):
        logging.error(
            f"\tat data row #{vars['input_row_number']}: "
            + colored(vars["_values"], colors.darkgray)
        )
    raise exception from None


def user_warning(message, exception=None, code=None):
    """
    Reports a warning.
    Prints a custom message.
    If `error_on_warning` is True, raises an error instead.
    """
    if error_on_warning:
        user_error(message, exception if exception else Exception(message), code)
    logging.warning(mk_user_msg("WARNING", colors.lightyellow, message, code))


def user_info(message, code=None):
    """
    Reports (verbose) information.
    """
    logging.info(mk_user_msg("INFO", colors.green, message, code, colors.yellow))


def user_debug(message, code=None):
    """
    Reports (verbose) information.
    """
    logging.debug(
        mk_user_msg("DEBUG", colors.lightcyan, message, code, colors.lightgray)
    )


def user_debug_dict(message, adict):
    """
    Reports (debug) information, printing a dict as a pretty json.
    """
    user_debug(message, json.dumps(adict, indent=4))


def user_warning4func(message, exception, *args, **kwargs):
    user_warning(
        message,
        exception,
        ",".join(
            [quote_ifstr(v) for v in args]
            + [f"{k}={quote_ifstr(v)}" for k, v in kwargs.items()]
        ),
    )


def conversion_warning(atype, exception, *args, **kwargs):
    user_warning4func(
        f"could not convert string to {atype}, returning NULL",
        exception,
        *args,
        **kwargs,
    )
